/**************************************************
 * VRML 2.0, Draft 2 Parser
 * Copyright (C) 1996 Silicon Graphics, Inc.
 *
 * Author(s)	: Gavin Bell
 *                Daniel Woods (first port)
 **************************************************
 */
%{
#include "tokens.h"
#include <string.h>

	/* Current line number */
int currentLineNumber = 1;
void yyResetLineNumber() { currentLineNumber = 1; }

extern void yyerror(const char *);

	/* The YACC parser sets this to a token to direct the lexer */
    /* in cases where just syntax isn't enough: */
int expectToken = 0;

	/* True when parsing a multiple-valued field: */
static int parsing_mf = 0;

	/* These are used when parsing SFImage fields: */
static int sfImageIntsParsed = 0;
static int sfImageIntsExpected = 0;

int yywrap() { BEGIN INITIAL; return 1; }

%}

	/* Normal state:  parsing nodes.  The initial start state is used */
    /* only to recognize the VRML header. */
%x NODE

	/* Start tokens for all of the field types, */
    /* except for MFNode and SFNode, which are almost completely handled */
    /* by the parser: */
%x SFB SFC SFF SFIMG SFI SFR SFS SFT SFV2 SFV3
%x MFC MFF MFI MFR MFS MFV2 MFV3
%x IN_SFS IN_MFS IN_SFIMG

	/* Big hairy expression for floating point numbers: */
float (-?(([0-9]+)|([0-9]*\.[0-9]+)([eE][+\-]?[0-9]+)?))

	/* Ints are decimal or hex (0x##): */
int (-?([0-9]+)|(0[xX][0-9a-fA-F]*))

	/* Whitespace.  Using this pattern can screw up currentLineNumber, */
    /* so it is only used wherever it is really convenient and it is */
    /* extremely unlikely that the user will put in a carriage return */
    /* (example: between the floats in an SFVec3f) */
ws ([ \t\r\n,]|(#.*))+
	/* And the same pattern without the newline */
wsnnl ([ \t\r,]|(#.*))

	/* Legal characters to start an identifier */
idStartChar ([^\x30-\x39\x00-\x20\x22\x23\x27\x2c\x2e\x5b-\x5d\x7b\x7d])
	/* Legal other characters in an identifier */
idRestChar	([^\x00-\x20\x22\x23\x27\x2c\x2e\x5b-\x5d\x7b\x7d])
%%

%{
	/* Switch into a new start state if the parser */
    /* just told us that we've read a field name */
    /* and should expect a field value (or IS) */
	if (expectToken != 0) {
      if (yy_flex_debug)
          fprintf(stderr,"LEX--> Start State %d\n", expectToken);
      
      /*
       * Annoying.  This big switch is necessary because
       * LEX wants to assign particular numbers to start
       * tokens, and YACC wants to define all the tokens
       * used, too.  Sigh.
       */
      switch(expectToken) {
       	case SFBOOL: BEGIN SFB; break;
        case SFCOLOR: BEGIN SFC; break;
        case SFFLOAT: BEGIN SFF; break;
        case SFIMAGE: BEGIN SFIMG; break;
        case SFINT32: BEGIN SFI; break;
        case SFROTATION: BEGIN SFR; break;
        case SFSTRING: BEGIN SFS; break;
        case SFTIME: BEGIN SFT; break;
        case SFVEC2F: BEGIN SFV2; break;
        case SFVEC3F: BEGIN SFV3; break;
        case MFCOLOR: BEGIN MFC; break;
        case MFFLOAT: BEGIN MFF; break;
        case MFINT32: BEGIN MFI; break;
        case MFROTATION: BEGIN MFR; break;
        case MFSTRING: BEGIN MFS; break;
        case MFVEC2F: BEGIN MFV2; break;
        case MFVEC3F: BEGIN MFV3; break;

        /* SFNode and MFNode are special.  Here the lexer just returns */
        /* "marker tokens" so the parser knows what type of field is */
        /* being parsed; unlike the other fields, parsing of SFNode/MFNode */
        /* field happens in the parser. */
        case MFNODE: expectToken = 0; return MFNODE;
        case SFNODE: expectToken = 0; return SFNODE;
        
        default: yyerror("ACK: Bad expectToken"); break;
      }
    }
%}

	/* These are more complicated than they really need to be because */
    /* I was ambitious and made the whitespace-matching rules aggressive */
<INITIAL>"#VRML V2.0 utf8".*\n{wsnnl}*	{
					  			  yyerror("VRML 2.0 final header found\n"
                      		      "This tool based on VRML Draft #2\n");
                                  BEGIN NODE;
                                }
<INITIAL>"#VRML Draft #2 V2.0 utf8".*\n{wsnnl}*		{ BEGIN NODE; }
<INITIAL>.*			{ yyerror("Not a VRML Draft #2 file: bad header\n"
							  "#VRML Draft #2 V2.0 utf8  is correct header");
                      return 0;
                    }

	/* The lexer is in the NODE state when parsing nodes, either */
    /* top-level nodes in the .wrl file, in a prototype implementation, */
    /* or when parsing the contents of SFNode or MFNode fields. */
<NODE>PROTO			{ return PROTO; }
<NODE>EXTERNPROTO	{ return EXTERNPROTO; }
<NODE>DEF			{ return DEF; }
<NODE>USE			{ return USE; }
<NODE>TO			{ return TO; }
<NODE>IS			{ return IS; }
<NODE>ROUTE			{ return ROUTE; }
<NODE>NULL			{ return SFN_NULL; }
<NODE>eventIn		{ return EVENTIN; }
<NODE>eventOut		{ return EVENTOUT; }
<NODE>field			{ return FIELD; }
<NODE>exposedField	{ return EXPOSEDFIELD; }

	/* Legal identifiers: */
<NODE>{idStartChar}{idRestChar}*	{ yylval.string = strdup(yytext);
									  return IDENTIFIER; }

	/* All fields may have an IS declaration: */
<SFB,SFC,SFF,SFIMG,SFI,SFR,SFS,SFT,SFV2,SFV3>IS		{ BEGIN NODE;
													  expectToken = 0;
                                                      yyless(0);
                                                    }
<MFC,MFF,MFI,MFR,MFS,MFV2,MFV3>IS     { BEGIN NODE;
									    expectToken = 0;
                                        yyless(0); /* put back the IS */
                                      }

    /* All MF field types other than MFNode are completely parsed here */
    /* in the lexer, and one token is returned to the parser.  They all */
    /* share the same rules for open and closing brackets: */
<MFC,MFF,MFI,MFR,MFS,MFV2,MFV3>\[ 	{ if (parsing_mf) yyerror("Double [");
									  parsing_mf = 1;
                                    }
<MFC,MFF,MFI,MFR,MFS,MFV2,MFV3>\] 	{ if (!parsing_mf) yyerror("Unmatched ]");
									  int fieldType = expectToken;
									  BEGIN NODE;
                                      parsing_mf = 0;
                                      expectToken = 0;
                                      return fieldType;
                                    }
                                      
<SFB>TRUE		{ BEGIN NODE; expectToken = 0; return SFBOOL; }
<SFB>FALSE		{ BEGIN NODE; expectToken = 0; return SFBOOL; }

<SFI>{int}		{ BEGIN NODE; expectToken = 0; return SFINT32; }
<MFI>{int}		{ if (parsing_mf) ;
					/* A real implementation would add the float to an */
                    /* array...  but here we do nothing */
				  else {
                  	  BEGIN NODE; expectToken = 0; return MFINT32;
                  }
                }

	/* All the floating-point types are pretty similar: */
<SFF>{float}		{ BEGIN NODE; expectToken = 0; return SFFLOAT; }
<MFF>{float}		{ if (parsing_mf) ; /* Add to array... */
                  	  else {
                      	/* No open bracket means a single value: */
                        BEGIN NODE; expectToken = 0; return MFFLOAT;
                      }
                    }
<SFV2>{float}{ws}{float}    { BEGIN NODE; expectToken = 0; return SFVEC2F; }
<MFV2>{float}{ws}{float}	{ if (parsing_mf) ; /* .. add to array... */
              		  		  else {
                              	BEGIN NODE; expectToken = 0; return MFVEC2F;
                              }
                            }
<SFV3>({float}{ws}){2}{float}	{ BEGIN NODE; expectToken = 0; return SFVEC3F; }
<MFV3>({float}{ws}){2}{float}	{ if (parsing_mf) ; /* .. add to array... */
              					  else {
                                    BEGIN NODE; expectToken = 0; return MFVEC3F;
                                  }
                                }
<SFR>({float}{ws}){3}{float}    { BEGIN NODE; expectToken = 0; return SFROTATION; }
<MFR>({float}{ws}){3}{float}	{ if (parsing_mf) ; /* .. add to array... */
			              		  else {
                                    BEGIN NODE; expectToken = 0; return MFROTATION;
                                  }
                                }
<SFC>({float}{ws}){2}{float}    { BEGIN NODE; expectToken = 0; return SFCOLOR; }
<MFC>({float}{ws}){2}{float}	{ if (parsing_mf) ; /* .. add to array... */
			              		  else {
                                    BEGIN NODE; expectToken = 0; return MFCOLOR;
                                  }
                                }
<SFT>{float}	    { BEGIN NODE; expectToken = 0; return SFTIME; }
               
    /* SFString/MFString */
<SFS,MFS>\"				{ BEGIN IN_SFS; }
	/* Anything besides open-quote (or whitespace) is an error: */
<SFS>[^ \"\t\r\,\n]+	{ yyerror("String missing open-quote");
						  BEGIN NODE; expectToken = 0; return SFSTRING;
                        }
    /* Expect open-quote, open-bracket, or whitespace: */
<MFS>[^ \[\]\"\t\r\,\n]+	{ yyerror("String missing open-quote");
						  	  BEGIN NODE; expectToken = 0; return MFSTRING;
                            }
    /* Backslashed-quotes are OK: */
<IN_SFS,IN_MFS>\\\"	   	;
    /* Gobble up anything besides quotes and newlines. */
    /* Newlines are legal in strings, but we exclude them here so */
    /* that line number are counted correctly by the catch-all newline */
    /* rule that applies to everything. */
<IN_SFS,IN_MFS>[^\"\n]+	;
	/* Quote ends the string: */
<IN_SFS>\"				{ BEGIN NODE; expectToken = 0; return SFSTRING; }
<IN_MFS>\"				{ if (parsing_mf) ; /* ... add to array ... */
						  else {
                          	  BEGIN NODE; expectToken = 0; return MFSTRING;
                          }
                        }

    /* SFImage: width height numComponents then width*height integers: */
<SFIMG>{int}{ws}{int}	{ int w, h;
						  sscanf(yytext, "%d %d", &w, &h);
                          sfImageIntsExpected = 1+w*h;
                          sfImageIntsParsed = 0;                          
                          BEGIN IN_SFIMG;
                        }
<IN_SFIMG>{int}         { ++sfImageIntsParsed;
						  if (sfImageIntsParsed == sfImageIntsExpected) {
                          	  BEGIN NODE; expectToken = 0; return SFIMAGE;
                          }
                        }

	/* Whitespace and catch-all rules apply to all start states: */
<*>{wsnnl}+		;
	/* This is also whitespace, but we'll keep track of line number */
    /* to report in errors: */
<*>{wsnnl}*\n{wsnnl}*		{ ++currentLineNumber; }

	/* This catch-all rule catches anything not covered by any of */
    /* the above: */
<*>. 			{ return yytext[0]; }

