//*******************************************************
// VRML 2.0, Draft 2 Parser
// Copyright (C) 1996 Silicon Graphics, Inc.
//
// Author(s)    : Gavin Bell
//                Daniel Woods (first port, minor fixes)
//*******************************************************
//
%{

//
// Parser for VRML 2.0 files.
// This is a minimal parser that does NOT generate an in-memory scene graph.
// 

// The original parser was developed on a Windows 95 PC with
// Borland's C++ 5.0 development tools.  This was then ported
// to a Windows 95 PC with Microsoft's MSDEV C++ 4.0 development
// tools.  The port introduced the ifdef's for
//    USING_BORLAND_CPP_5          : since this provides a "std namespace",
//    TWO_ARGUMENTS_FOR_STL_STACK  : STL is a moving target.  The stack template
//                                     class takes either one or two arguments.

#include <iostream.h>
#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include <list.h>
#include <stack.h>
#include <string.h>

#include "VrmlNodeType.h"

#define YYDEBUG 1

#ifdef USING_BORLAND_CPP_5
using namespace std;
#endif

// Currently-being-define proto.  Prototypes may be nested, so a stack
// is needed:

#ifdef TWO_ARGUMENTS_FOR_STL_STACK  
stack< VrmlNodeType*, list<VrmlNodeType*> > currentProtoStack;
#else
stack< list<VrmlNodeType*> > currentProtoStack;
#endif  /* TWO_ARGUMENTS_FOR_STL_STACK */

// This is used to keep track of which field in which type of node is being
// parsed.  Field are nested (nodes are contained inside MFNode/SFNode fields)
// so a stack of these is needed:
typedef struct {
	const VrmlNodeType *nodeType;
    const char *fieldName;
} FieldRec;
#ifdef TWO_ARGUMENTS_FOR_STL_STACK  
stack< FieldRec*, list<FieldRec*> > currentField;
#else
stack< list<FieldRec*> > currentField;
#endif  /* TWO_ARGUMENTS_FOR_STL_STACK */

// This is used when the parser knows what kind of token it expects
// to get next-- used when parsing field values (whose types are declared
// and read by the parser) and at certain other places:
extern int expectToken;

// Current line number (set by lexer)
extern int currentLineNumber;

// Some helper routines defined below:
void beginProto(const char *);
void endProto();
int addField(const char *type, const char *name);
int addEventIn(const char *type, const char *name);
int addEventOut(const char *type, const char *name);
int addExposedField(const char *type, const char *name);
int add(void (VrmlNodeType::*)(const char *, int), const char *, const char *);
int fieldType(const char *type);
void enterNode(const char *);
void exitNode();
void inScript();
void enterField(const char *);
void exitField();
void expect(int type);

%}

%union {
	char *string;

    /* Other types that will be needed by a true VRML implementation
     * (but are not defined by this parser due to the complexity):
     * Node *node;
     * list<Node *> *nodeList;
     */
};

%type <string> IDENTIFIER

/*
 * And types that will be needed by a true VRML implementation:
 * %type <nodeList> vrmlscene declarations
 */

%token IDENTIFIER DEF USE PROTO EXTERNPROTO TO IS ROUTE SFN_NULL
%token EVENTIN EVENTOUT FIELD EXPOSEDFIELD
%token SFBOOL SFCOLOR SFFLOAT SFIMAGE SFINT32 SFNODE SFROTATION
%token SFSTRING SFTIME SFVEC2F SFVEC3F
%token MFCOLOR MFFLOAT MFINT32 MFROTATION MFSTRING MFVEC2F MFVEC3F
%token MFNODE

%%

vrmlscene:	declarations
	;

declarations:
		/* Empty is OK */
	|	declarations declaration
    ;

declaration:
		nodeDeclaration
	|	protoDeclaration
	|	routeDeclaration
    ;

nodeDeclaration:
		node
	|	DEF IDENTIFIER node		{ free($2); }
    |	USE IDENTIFIER			{ free($2); }
    ;

protoDeclaration:
		proto
	|	externproto
    ;

proto:
		 PROTO IDENTIFIER			{ beginProto($2); }						
	     	'[' interfaceDeclarations ']'
            '{' declarations '}'	{ endProto();  free($2);}
	;

externproto:
		 EXTERNPROTO IDENTIFIER			{ beginProto($2); }						
	     	'[' externInterfaceDeclarations ']' 
            							{ expect(MFSTRING); }
            fieldValue					{ endProto();  free($2); }
	;
interfaceDeclarations:
		/* Empty is OK */
	|	interfaceDeclarations interfaceDeclaration
    ;

interfaceDeclaration:
		EVENTIN IDENTIFIER IDENTIFIER		{ addEventIn($2, $3);
        									  free($2); free($3); }
    |   EVENTOUT IDENTIFIER IDENTIFIER		{ addEventOut($2, $3);
        									  free($2); free($3); }
    |   FIELD IDENTIFIER IDENTIFIER			{ int type = addField($2, $3);
    										  expect(type); }
    	    fieldValue        				{ free($2); free($3); }
	|	EXPOSEDFIELD IDENTIFIER	IDENTIFIER	{ int type = addExposedField($2, $3);
    										  expect(type); }
    		fieldValue                      { free($2); free($3); }
    ;

externInterfaceDeclarations:
		/* Empty is OK */
     |  externInterfaceDeclarations externInterfaceDeclaration
     ;

externInterfaceDeclaration:
		EVENTIN IDENTIFIER IDENTIFIER		{ addEventIn($2, $3);
        									  free($2); free($3); }
    |   EVENTOUT IDENTIFIER IDENTIFIER		{ addEventOut($2, $3);
        									  free($2); free($3); }
    |	FIELD IDENTIFIER IDENTIFIER			{ addField($2, $3);
        									  free($2); free($3); }
    |	EXPOSEDFIELD IDENTIFIER IDENTIFIER	{ addExposedField($2, $3);
        									  free($2); free($3); }
    ;

routeDeclaration:
		ROUTE IDENTIFIER '.' IDENTIFIER TO IDENTIFIER '.' IDENTIFIER
        		{ free($2); free($4); free($6); free($8); }
    ; 

node:
		IDENTIFIER  				{ enterNode($1); }
        	'{' nodeGuts '}'		{ exitNode(); free($1);}
    ;

nodeGuts:
		/* Empty is OK */
    | 	nodeGuts nodeGut
    ;

nodeGut:
		IDENTIFIER 					{ enterField($1); }
        		fieldValue			{ exitField(); free($1); }
    |	routeDeclaration
    |	protoDeclaration

    	/* The following are only valid for Script nodes: */
	|   EVENTIN IDENTIFIER IDENTIFIER		{ inScript(); free($2); free($3); }
    |   EVENTOUT IDENTIFIER IDENTIFIER		{ inScript(); free($2); free($3); }
    |   FIELD IDENTIFIER IDENTIFIER			{ inScript(); 
    										  int type = fieldType($2);
    										  expect(type); }
    	    fieldValue                      { free($2); free($3); }
    |	EVENTIN IDENTIFIER IDENTIFIER IS IDENTIFIER
    			{ inScript(); free($2); free($3); free($5); }
    |	EVENTOUT IDENTIFIER IDENTIFIER IS IDENTIFIER
    			{ inScript(); free($2); free($3); free($5); }
   ;

fieldValue:
		SFBOOL
    |   SFCOLOR
    |	MFCOLOR
    |   SFFLOAT
    |	MFFLOAT
    |	SFIMAGE
    |	SFINT32
    |	MFINT32
    |	SFROTATION
    |	MFROTATION
    |	SFSTRING
    |	MFSTRING
    |	SFTIME
    |	SFVEC2F
    |	MFVEC2F
    |	SFVEC3F
    |	MFVEC3F
    
    |	SFNODE nodeDeclaration
    |	SFNODE SFN_NULL
    |	MFNODE mfnodeValue
    |	IS IDENTIFIER				{ free($2); }
    ;

mfnodeValue:
		'[' nodes ']'
     |  nodeDeclaration
     ;

nodes:
		/* Empty is OK */
     |  nodes nodeDeclaration
     ;

%%

void
yyerror(const char *msg)
{
	cerr << "Error near line " << currentLineNumber << ": " << msg << "\n";
    expect(0);
}

static void
beginProto(const char *protoName)
{
	// Any protos in the implementation are in a local namespace:
    VrmlNodeType::pushNameSpace();

	VrmlNodeType *t = new VrmlNodeType(protoName);
    currentProtoStack.push(t);
}

static void
endProto()
{
	// Make any protos defined in implementation unavailable:
	VrmlNodeType::popNameSpace();

    // Add this proto definition:
	if (currentProtoStack.empty()) {
    	cerr << "Error: Empty PROTO stack!\n";
    }
    else {
    	VrmlNodeType *t = currentProtoStack.top();
    	currentProtoStack.pop();
        VrmlNodeType::addToNameSpace(t);
    }
}

int
addField(const char *type, const char *name)
{
	return add(&VrmlNodeType::addField, type, name);
}

int
addEventIn(const char *type, const char *name)
{
	return add(&VrmlNodeType::addEventIn, type, name);
}
int
addEventOut(const char *type, const char *name)
{
	return add(&VrmlNodeType::addEventOut, type, name);
}
int
addExposedField(const char *type, const char *name)
{
	return add(&VrmlNodeType::addExposedField, type, name);
}

int
add(void (VrmlNodeType::*func)(const char *, int), 
	const char *typeString, const char *name)
{
	int type = fieldType(typeString);

    if (type == 0) {
    	cerr << "Error: invalid field type: " << type << "\n";
    }

    // Need to add support for Script nodes:
	// if (inScript) ... ???

	if (currentProtoStack.empty()) {
    	cerr << "Error: declaration outside of prototype\n";
        return 0;
    }
    VrmlNodeType *t = currentProtoStack.top();
    (t->*func)(name, type);

    return type;
}

int
fieldType(const char *type)
{
    if (strcmp(type, "SFBool") == 0) return SFBOOL;
    if (strcmp(type, "SFColor") == 0) return SFCOLOR;
    if (strcmp(type, "SFFloat") == 0) return SFFLOAT;
    if (strcmp(type, "SFImage") == 0) return SFIMAGE;
    if (strcmp(type, "SFInt32") == 0) return SFINT32;
    if (strcmp(type, "SFNode") == 0) return SFNODE;
    if (strcmp(type, "SFRotation") == 0) return SFROTATION;
    if (strcmp(type, "SFString") == 0) return SFSTRING;
    if (strcmp(type, "SFTime") == 0) return SFTIME;
    if (strcmp(type, "SFVec2f") == 0) return SFVEC2F;
    if (strcmp(type, "SFVec3f") == 0) return SFVEC3F;
    if (strcmp(type, "MFColor") == 0) return MFCOLOR;
    if (strcmp(type, "MFFloat") == 0) return MFFLOAT;
    if (strcmp(type, "MFInt32") == 0) return MFINT32;
    if (strcmp(type, "MFNode") == 0) return MFNODE;
    if (strcmp(type, "MFRotation") == 0) return MFROTATION;
    if (strcmp(type, "MFString") == 0) return MFSTRING;
    if (strcmp(type, "MFVec2f") == 0) return MFVEC2F;
    if (strcmp(type, "MFVec3f") == 0) return MFVEC3F;

    cerr << "Illegal field type: " << type << "\n";

    return 0;
}

void
enterNode(const char *nodeType)
{
	const VrmlNodeType *t = VrmlNodeType::find(nodeType);
    if (t == NULL) {
    	char tmp[1000];
        sprintf(tmp, "Unknown node type '%s'", nodeType);
    	yyerror(tmp);
    }
    FieldRec *fr = new FieldRec;
    fr->nodeType = t;
    fr->fieldName = NULL;
    currentField.push(fr);
}

void
exitNode()
{
	FieldRec *fr = currentField.top();
    assert(fr != NULL);
	currentField.pop();

    delete fr;
}

void
inScript()
{
	FieldRec *fr = currentField.top();
    if (fr->nodeType == NULL ||
    	strcmp(fr->nodeType->getName(), "Script") != 0) {
        yyerror("interface declaration outside of Script or prototype");
    }
}

void
enterField(const char *fieldName)
{
	FieldRec *fr = currentField.top();
    assert(fr != NULL);

    fr->fieldName = fieldName;
    if (fr->nodeType != NULL) {
    	// enterField is called when parsing eventIn and eventOut IS
        // declarations, in which case we don't need to do anything special--
        // the IS IDENTIFIER will be returned from the lexer normally.
    	if (fr->nodeType->hasEventIn(fieldName) ||
        	fr->nodeType->hasEventOut(fieldName))
            return;
    
        int type = fr->nodeType->hasField(fieldName);
        if (type != 0) {
        	// Let the lexer know what field type to expect:
        	expect(type);
        }
        else {
        	cerr << "Error: Node's of type " << fr->nodeType->getName() <<
            	" do not have fields/eventIn/eventOut named " <<
                fieldName << "\n";
            // expect(ANY_FIELD);
        }
    }
    // else expect(ANY_FIELD);
}

void
exitField()
{
	FieldRec *fr = currentField.top();
    assert(fr != NULL);

	fr->fieldName = NULL;
}

void
expect(int type)
{
	expectToken = type;
}

