/**************************************************
 * VRML 2.0, Draft 2 Parser
 * Copyright (C) 1996 Silicon Graphics, Inc.
 *
 * Author(s)    : Gavin Bell
 *                Daniel Woods (first port)
 **************************************************
 */

//
// The VrmlNodeType class is responsible for storing information about node
// or prototype types.
//

#include "VrmlNodeType.h"
#include <iostream.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

//
// Static list of node types.
//
list<VrmlNodeType*> VrmlNodeType::typeList;

VrmlNodeType::VrmlNodeType(const char *nm)
{
	assert(nm != NULL);
	name = strdup(nm);
}

VrmlNodeType::~VrmlNodeType()
{
	free(name);

    // Free strings duplicated when fields/eventIns/eventOuts added:
    list<NameTypeRec*>::iterator i;

    for (i = eventIns.begin(); i != eventIns.end(); i++) {
    	NameTypeRec *r = *i;
    	free(r->name);
        delete r;
    }
    for (i = eventOuts.begin(); i != eventOuts.end(); i++) {
    	NameTypeRec *r = *i;
    	free(r->name);
        delete r;
    }
    for (i = fields.begin(); i != fields.end(); i++) {
    	NameTypeRec *r = *i;
    	free(r->name);
        delete r;
    }
}

void
VrmlNodeType::addToNameSpace(VrmlNodeType *_type)
{
	if (find(_type->getName()) != NULL) {
    	cerr << "PROTO " << _type->getName() << " already defined\n";
        return;
    }
    typeList.push_front(_type);
}

//
// One list is used to store all the node types.  Nested namespaces are
// separated by NULL elements.
// This isn't terribly efficient, but it is nice and simple.
//
void
VrmlNodeType::pushNameSpace()
{
	typeList.push_front(NULL);
}

void
VrmlNodeType::popNameSpace()
{
	// Remove everything up to and including the next NULL marker:
    list<VrmlNodeType*>::iterator i;
    for (i = typeList.begin(); i != typeList.end();) {
    	VrmlNodeType *nodeType = *i;
        ++i;
        typeList.pop_front();

        if (nodeType == NULL) {
            break;
        }
        else {
            // NOTE:  Instead of just deleting the VrmlNodeTypes, you will
            // probably want to reference count or garbage collect them, since
            // any nodes created as part of the PROTO implementation will
            // probably point back to their VrmlNodeType structure.
            delete nodeType;
        }
    }
}

const VrmlNodeType *
VrmlNodeType::find(const char *_name)
{
	// Look through the type stack:
    list<VrmlNodeType*>::iterator i;
    for (i = typeList.begin(); i != typeList.end(); i++) {
    	const VrmlNodeType *nt = *i;
    	if (nt != NULL && strcmp(nt->getName(),_name) == 0) {
            return nt;
        }
    }
    return NULL;
}

void
VrmlNodeType::addEventIn(const char *name, int type)
{
	add(eventIns, name, type);
};
void
VrmlNodeType::addEventOut(const char *name, int type)
{
	add(eventOuts, name, type);
};
void
VrmlNodeType::addField(const char *name, int type)
{
	add(fields, name, type);
};
void
VrmlNodeType::addExposedField(const char *name, int type)
{
	char tmp[1000];
	add(fields, name, type);
    sprintf(tmp, "set_%s", name);
    add(eventIns, tmp, type);
    sprintf(tmp, "%s_changed", name);
    add(eventOuts, tmp, type);
};

void
VrmlNodeType::add(list<NameTypeRec*> &recs, const char *name, int type)
{
	NameTypeRec *r = new NameTypeRec;
    r->name = strdup(name);
    r->type = type;
	recs.push_front(r);
}

int
VrmlNodeType::hasEventIn(const char *name) const
{
	return has(eventIns, name);
}
int
VrmlNodeType::hasEventOut(const char *name) const
{
	return has(eventOuts, name);
}
int
VrmlNodeType::hasField(const char *name) const
{
	return has(fields, name);
}
int
VrmlNodeType::hasExposedField(const char *name) const
{
	// Must have field "name", eventIn "set_name", and eventOut
    // "name_changed", all with same type:
	char tmp[1000];
	int type;
	if ( (type = has(fields, name)) == 0) return 0;

    sprintf(tmp, "set_%s\n", name);
    if (type != has(eventIns, name)) return 0;

    sprintf(tmp, "%s_changed", name);
    if (type != has(eventOuts, name)) return 0;

    return type;
}
int
VrmlNodeType::has(const list<NameTypeRec*> &recs, const char *name) const
{
	list<NameTypeRec*>::const_iterator i;
    for (i = recs.begin(); i != recs.end(); i++) {
    	if (strcmp((*i)->name, name) == 0)
        	return (*i)->type;
    }
    return 0;
}

